unit notefile;

interface

uses
	Contnrs, Classes;

const
	MaxLenObjName = 30;

resourcestring
	sAppTitle = 'Navision Object Text Explorer';

type
  TObjectTextFile = class;

	TObjectType = (otUnknown, otTable, otForm, otReport, otDataport, otCodeunit, otXMLPort, otMenuSuite);
	TObjectID = 1..MaxInt;
	TObjectName = string[MaxLenObjName];

  TObjectTextObject = class(TObject)
  private
    FObjectID: TObjectID;
    FObjectName: TObjectName;
    FObjectType: TObjectType;
    FOffset: integer;
    FParent: TObjectTextFile;
    FSize: integer;
  protected
    function GetDefaultFileName: string; virtual;
    function GetSignature: string; virtual;
    function GetText: string; virtual;
    procedure SaveToStream(Stream: TStream); virtual;
	public
    constructor Create(Parent: TObjectTextFile; Offset: integer; Signature: string);
    function FindText(var Offset: integer; Text: string; MatchCase: boolean): boolean; virtual;
    procedure ParseSignature(Signature: string); virtual;
    procedure SaveToFile(const FileName: string); virtual;
    procedure SetEnd(Value: integer); virtual;
    property DefaultFileName: string read GetDefaultFileName;
    property ObjectID: TObjectID read FObjectID;
    property ObjectName: TObjectName read FObjectName;
    property ObjectType: TObjectType read FObjectType;
    property Offset: integer read FOffset;
    property Parent: TObjectTextFile read FParent;
    property Signature: string read GetSignature;
    property Size: integer read FSize;
    property Text: string read GetText;
  end;

  TProgressStage = (psStarting, psRunning, psEnding);
  TProgressEvent = procedure(Sender: TObject; Stage: TProgressStage; PercentDone: byte) of object;

  TObjectTextFile = class(TObject)
  private
    FStream: TStream;
    FFileName: string;
    FObjects: TObjectList;
    FOnProgress: TProgressEvent;
  protected
    procedure DoProgress(Stage: TProgressStage; PercentDone: byte); virtual;
    function GetCount: integer; virtual;
    function GetObject(Index: integer): TObjectTextObject; virtual;
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure SetTextStr(const Value: string); virtual;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear; virtual;
    function FindText(var ObjectIndex: integer; var Offset: integer; Text: string; MatchCase: boolean): boolean; virtual;
    procedure LoadFromFile(const FileName: string); virtual;
    property Count: integer read GetCount;
    property FileName: string read FFileName;
    property Objects[Index: integer]: TObjectTextObject read GetObject;
    property OnProgress: TProgressEvent read FOnProgress write FOnProgress;
  end;

resourcestring
  sUnknown = 'Unknown';
	sTable = 'Table';
	sForm = 'Form';
	sReport = 'Report';
	sDataport = 'Dataport';
	sCodeunit = 'Codeunit';
	sXMLPort = 'XMLport';
	sMenuSuite = 'MenuSuite';

const
	sObjectType: array[TObjectType] of string = (sUnknown, sTable, sForm, sReport, sDataport, sCodeunit, sXMLPort, sMenuSuite);

implementation

uses
	StrUtils,
	SysUtils;

{ TObjectTextObject }

constructor TObjectTextObject.Create(Parent: TObjectTextFile; Offset: integer;
	Signature: string);
begin
	inherited Create;
	FOffset := Offset;
	FParent := Parent;
	ParseSignature(Signature);
end;

function TObjectTextObject.FindText(var Offset: integer; Text: string; MatchCase: boolean): boolean;
var
	ObjectText: string;
begin
	ObjectText := GetText;

	if not MatchCase then
	begin
		Text := Uppercase(Text);
		ObjectText := Uppercase(ObjectText);
	end;

	Offset := PosEx(Text, ObjectText, Offset);
	Result := (Offset <> 0);
end;

function TObjectTextObject.GetDefaultFileName: string;
begin
  Result := Format('%s%d.txt', [Lowercase(Copy(sObjectType[ObjectType], 1, 3)), ObjectID]);
end;

function TObjectTextObject.GetSignature: string;
begin
  Result := Format('%s %d %s', [sObjectType[ObjectType], ObjectID, ObjectName]);
end;

function TObjectTextObject.GetText: string;
begin
  SetString(Result, nil, Size);
  Parent.FStream.Seek(Offset, soBeginning);
  Parent.FStream.Read(Pointer(Result)^, Size);
end;

procedure TObjectTextObject.ParseSignature(Signature: string);
var
  FromChar, ToChar: integer;
begin
  ToChar := 0;

  repeat
    Inc(ToChar);
  until Signature[ToChar] = ' ';

	case Signature[1] of
    'T': FObjectType := otTable;
    'F': FObjectType := otForm;
    'R': FObjectType := otReport;
		'D': FObjectType := otDataport;
		'C': FObjectType := otCodeunit;
		'X': FObjectType := otXMLPort;
		'M': FObjectType := otMenuSuite;
	end;

	FromChar := ToChar + 1;

	repeat
		Inc(ToChar)
	until Signature[ToChar] = ' ';

	FObjectID := StrToInt(Copy(Signature, FromChar, ToChar - FromChar));
	FObjectName := Copy(Signature, ToChar, Length(Signature) - ToChar + 1);
end;

procedure TObjectTextObject.SaveToFile(const FileName: string);
var
	Stream: TStream;
begin
	Stream := TFileStream.Create(FileName, fmCreate);
	try
		SaveToStream(Stream);
	finally
		Stream.Free;
	end;
end;

procedure TObjectTextObject.SaveToStream(Stream: TStream);
var
	S: string;
begin
	S := GetText;
	Stream.WriteBuffer(Pointer(S)^, Length(S));
end;

procedure TObjectTextObject.SetEnd(Value: integer);
begin
  FSize := Value - Offset;
end;

{ TObjectTextFile }

constructor TObjectTextFile.Create;
begin
	inherited;
  FObjects := TObjectList.Create;
end;

destructor TObjectTextFile.Destroy;
begin
  FStream.Free;
  FObjects.Free;
  inherited;
end;

procedure TObjectTextFile.DoProgress(Stage: TProgressStage; PercentDone: byte);
begin
  if Assigned(OnProgress) then
    OnProgress(Self, Stage, PercentDone);
end;

function TObjectTextFile.GetCount: integer;
begin
  Result := FObjects.Count;
end;

function TObjectTextFile.GetObject(Index: integer): TObjectTextObject;
begin
  Result := FObjects[Index] as TObjectTextObject;
end;

procedure TObjectTextFile.LoadFromStream(Stream: TStream);
var
  Size: integer;
  S: string;
begin
  FObjects.Clear;

  Size := Stream.Size - Stream.Position;
  SetString(S, nil, Size);
  Stream.Read(Pointer(S)^, Size);
  SetTextStr(S);
end;

procedure TObjectTextFile.SetTextStr(const Value: string);
resourcestring
  sObject = 'OBJECT ';
var
  First, Current, Start: PChar;
  Line: String;
  PrevProgress, CurrProgress: integer;
  ObjectTextObject: TObjectTextObject;
begin
	ObjectTextObject := nil;
  DoProgress(psStarting, 0);

  PrevProgress := 0;
  First := Pointer(Value);
  Current := First;

  if Current <> nil then
    while (Current - First) < Length(Value) do
    begin
      Start := Current;
      while not (Current^ in [#0, #10, #13]) do Inc(Current);
      SetString(Line, Start, Current - Start);

      if Copy(Line, 1, Length(sObject)) = sObject then
      begin
        if Assigned(ObjectTextObject) then
          ObjectTextObject.SetEnd(Start - First);
        Delete(Line, 1, Length(sObject));
        ObjectTextObject := TObjectTextObject.Create(Self, Start - First, Line);
        FObjects.Add(ObjectTextObject);
      end
      else
      begin
        CurrProgress := Round((Current - First) / Length(Value) * 100);
        if CurrProgress <> PrevProgress then
        begin
          DoProgress(psRunning, CurrProgress);
          PrevProgress := CurrProgress;
        end;
      end;

      if Current^ = #13 then Inc(Current);
      if Current^ = #10 then Inc(Current);
    end;

  if Assigned(ObjectTextObject) then
    ObjectTextObject.SetEnd(Current - First);

  DoProgress(psEnding, 100);
end;

procedure TObjectTextFile.LoadFromFile(const FileName: string);
begin
  FStream.Free;
  FStream := TFileStream.Create(FileName, fmOpenRead or fmShareExclusive);
  LoadFromStream(FStream);
  FFileName := FileName;
end;

function TObjectTextFile.FindText(var ObjectIndex: integer; var Offset: integer;
  Text: string; MatchCase: boolean): boolean;
begin
  while ObjectIndex <= Count - 1 do
  begin
    if Objects[ObjectIndex].FindText(Offset, Text, MatchCase) then
    begin
      Result := true;
      exit;
    end;

    Inc(ObjectIndex);
    Offset := 0;
  end;

  Result := false;
end;

procedure TObjectTextFile.Clear;
begin
  FObjects.Clear;
  FFileName := '';
end;

end.
